<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:sch="http://purl.oclc.org/dsdl/schematron"
                xmlns:vlc=
          "http://docs.oasis-open.org/codelist/ns/ContextValueAssociation/1.0/"
                version="1.0"
                exclude-result-prefixes="vlc">
<!--
     An implementation in XSLT of the OASIS Code List TC's Schematron Methodology for Code List and Value Validation.

     A stylesheet to convert an agreed upon set of pointers to coded values
     for constraining instances of documents at particular XPath addresses.

     This stylesheet calls out to external template rules to accommodate the
     naming of metadata and the format of external code list enumerations,
     thus it is not hardwired to UBL or any external representation of values.

     It should not be necessary to modify this file in order to support
     any definition of business document meta data or value list vocabulary.

     Implementation note: literal result elements are not used in this fragment
     because of the desire to emit Schematron elements in the default namespace
     while needing the namespace declared for matching purposes.

     Implementation restriction:  at this time this script is checking
     individual code list association files for having non-conflicting prefixes
     for namespaces, but it is not checking the entire suite of prefixes for
     namespaces for every included file for unique usage of prefix values in
     the generated result.  This may go unreported and cause a problem in the
     Schematron processing.

  $Id: Constraints2Schematron.xsl,v 1.9 2007/09/01 19:52:33 Ram Kumar Exp $

OASIS Customer Information Quality (CIQ) Technical Committee

Copyright (C) OASIS Open (2006). All Rights Reserved.

This document and translations of it may be copied and furnished to others, and
derivative works that comment on or otherwise explain it or assist in its
implementation may be prepared, copied, published and distributed, in whole or
in part, without restriction of any kind, provided that the above copyright
notice and this paragraph are included on all such copies and derivative works.
However, this document itself may not be modified in any way, such as by
removing the copyright notice or references to OASIS, except as needed for the
purpose of developing OASIS specifications, in which case the procedures for
copyrights defined in the OASIS Intellectual Property Rights document must be
followed, or as required to translate it into languages other than English.

The limited permissions granted above are perpetual and will not be revoked by
OASIS or its successors or assigns.

This document and the information contained herein is provided on an "AS IS"
basis and OASIS DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO ANY WARRANTY THAT THE USE OF THE INFORMATION HEREIN WILL NOT
INFRINGE ANY RIGHTS OR ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE.

-->

<!--these shortcuts make some namespace work more succinct in the stylesheet-->
<xsl:variable name="vlc-namespace" select="document('')/*/namespace::vlc"/>
<xsl:variable name="sch-namespace" select="document('')/*/
                                                namespace::*[name(.)='']"/>

<!--this keeps code lists handy for reference-->
<xsl:key name="valuelists" match="vlc:ValueList" use="@xml:id"/>

<!--this remembers the association file root node for the initial file only-->
<xsl:variable name="assoc-file" select="document('',/)"/>
<xsl:variable name="assoc-uri"/>

<!--this makes the resulting file easy to edit-->
<xsl:output encoding="ISO-8859-1"/>

<xsl:template match="/">
  <xsl:param name="assoc-file" select="$assoc-file"/>
  <xsl:param name="assoc-uri" select="$assoc-uri"/>
  <!--prevent stylesheet abuse-->
  <xsl:if test="not($assoc-file/vlc:ValueListConstraints)">
    <xsl:message terminate="yes">
      <xsl:if test="$assoc-uri">
        <xsl:text>Problem processing file '</xsl:text>
        <xsl:value-of select="$assoc-uri"/>
        <xsl:text>': </xsl:text>
      </xsl:if>
      <xsl:text>Expected the document element {</xsl:text>
      <xsl:value-of select="$vlc-namespace"/>
      <xsl:text>}ValueListConstraints, but instead have {</xsl:text>
      <xsl:value-of select="concat( namespace-uri(*),'}',local-name(*) )"/>
    </xsl:message>
  </xsl:if>
  <xsl:if test="not($assoc-uri)">
    <!--first time around, check the include tree for infinite loops-->
    <xsl:call-template name="check-infinite-loops"/>
  </xsl:if>
  <!--check given instance for inconsistencies-->
  <xsl:call-template name="check-inconsistencies"/>
  <xsl:choose>
    <xsl:when test="$assoc-uri">
      <!--this has been invoked by an Include directive, so just add rules-->
      <xsl:call-template name="add-rules">
        <xsl:with-param name="assoc-file" select="$assoc-file"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <!--create the pattern-->
      <xsl:element name="pattern" 
                   namespace="http://purl.oclc.org/dsdl/schematron">
        <xsl:attribute name="id">
          <xsl:value-of select="$assoc-file/vlc:ValueListConstraints/@name"/>
        </xsl:attribute>
        <xsl:call-template name="add-rules">
          <xsl:with-param name="assoc-file" select="$assoc-file"/>
          <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
        </xsl:call-template>
      </xsl:element>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--========================================================================-->
<!--add rules to the result-->

<!--add the rules from the context association file to the result-->
<xsl:template name="add-rules">
  <xsl:param name="assoc-file" select="$assoc-file"/>
  <xsl:param name="assoc-uri" select="$assoc-uri"/>
  <xsl:text>
</xsl:text>
  <xsl:if test="not($assoc-uri)">
    <!--$comment must be supplied by an external fragment; not defined here-->
    <xsl:comment><xsl:value-of select="$comment"/></xsl:comment><xsl:text>
</xsl:text>
  </xsl:if>
  <xsl:comment>
    Start of synthesis of rules from code list context associations.
<xsl:if test="normalize-space($assoc-file/*/vlc:Title)">
  <!--this will help identify the sets of rules-->
  <xsl:value-of select="$assoc-file/*/vlc:Title"/>
  <xsl:text>
</xsl:text>
</xsl:if>
<xsl:if test="normalize-space($assoc-file/*/vlc:Identification)">
  <!--this will help identify the sets of rules-->
  <xsl:value-of select="$assoc-file/*/vlc:Identification"/>
  <xsl:text>
</xsl:text>
</xsl:if>
<!--determine end-user namespaces by ignoring expected namespaces-->
<xsl:variable name="ns" 
              select="$assoc-file//namespace::*[name(.) != 'xml' and 
                                . != $vlc-namespace and . != $sch-namespace]"/>
<xsl:if test="$ns">
    Required namespace declarations as indicated in this set of rules:

<xsl:for-each select="$ns">
  <xsl:if test="generate-id(.)=generate-id($ns[name(.)=name(current())][1])">
    <xsl:value-of select="concat('&lt;ns prefix=&quot;',name(.),
                          '&quot; uri=&quot;',.,'&quot;/>')"/>
      <xsl:text>
</xsl:text>
  </xsl:if>
</xsl:for-each>
      <xsl:text>
</xsl:text>
</xsl:if>
</xsl:comment><xsl:text>
</xsl:text>
  <xsl:apply-templates select="$assoc-file/*/vlc:Contexts/vlc:Context">
    <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
  </xsl:apply-templates>
  <xsl:comment>
    End of synthesis of rules from code list context associations.
</xsl:comment><xsl:text>
</xsl:text>
  <!--copy any indentation that was at the end of the rules-->
  <xsl:copy-of select="node()[last()][self::text()]"/>

  <xsl:for-each select="$assoc-file/*/vlc:Include">
    <xsl:sort select="position()" order="descending" data-type="number"/>
    <xsl:call-template name="add-rules">
      <xsl:with-param name="assoc-file" select="document(@uri,.)"/>
      <xsl:with-param name="assoc-uri" select="@uri"/>
    </xsl:call-template>
  </xsl:for-each>
</xsl:template>

<!--========================================================================-->
<!--create the rules; one rule for each context-->

<xsl:template match="vlc:Context" name="gather-context">
  <xsl:param name="assoc-uri"/>
  <xsl:param name="valuelist-references" 
             select="concat( normalize-space(@values), ' ' )"/>
  <xsl:param name="valuelists-referenced" select="/.."/>
  <!--process the context-->
  <xsl:choose>
    <xsl:when test="contains( $valuelist-references,' ' )">
      <!--gather up code list references for the given context-->
      <xsl:call-template name="gather-context">
        <xsl:with-param  name="valuelist-references"
                         select="substring-after($valuelist-references,' ')"/>
        <xsl:with-param  name="valuelists-referenced"
                         select="$valuelists-referenced | 
               key('valuelists',substring-before($valuelist-references,' '))"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <!--now have the union set of all referenced in given context-->
      <xsl:variable name="context">
        <!--assume the context is the item address, but xpath overrides it-->
        <xsl:choose>
          <xsl:when test="@xpath"><xsl:value-of select="@xpath"/></xsl:when>
          <xsl:when test="@context">
            <xsl:value-of select="concat(@context,'//',@item)"/>
          </xsl:when>
          <xsl:otherwise><xsl:value-of select="@item"/></xsl:otherwise>
        </xsl:choose>
      </xsl:variable>
      <!--make the rules for this context-->
      <xsl:call-template name="code-list-make-rule">
        <xsl:with-param name="item" select="@item"/>
        <xsl:with-param name="context" select="$context"/>
        <xsl:with-param name="valuelists-referenced" 
                        select="$valuelists-referenced"/>
        <xsl:with-param name="values-identifier" 
                        select="normalize-space(@values)"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--compose the report of a context violation-->
<xsl:template name="compose-violation-report">
  <xsl:param name="values-identifier"/>
  <xsl:param name="assoc-uri"/>
  <xsl:param name="context"/>
  <xsl:choose>
    <xsl:when test="normalize-space(vlc:Message) or vlc:Message/*">
      <!--the context includes the desired messages to be displayed-->
      <xsl:apply-templates mode="sch-only" select="vlc:Message/node()"/>
    </xsl:when>
    <xsl:otherwise>
      <!--a default message is composed to report the violation-->
      <xsl:text>Value supplied '</xsl:text>
      <xsl:element name="value-of" 
                   namespace="http://purl.oclc.org/dsdl/schematron">
        <xsl:attribute name="select">.</xsl:attribute>
      </xsl:element>
      <xsl:text>' is unacceptable for values identified by '</xsl:text>
      <xsl:value-of select="$values-identifier"/>
      <xsl:if test="$assoc-uri">
        <xsl:text>' in association file with uri '</xsl:text>
        <xsl:value-of select="$assoc-uri"/>
      </xsl:if>
      <xsl:text>' in the context '</xsl:text>
      <xsl:value-of select="$context"/>'<xsl:text/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--when copying the message content from CVA file, trim to Schematron ns-->
<xsl:template mode="sch-only" match="node()">
  <xsl:copy>
    <xsl:copy-of select="@*"/>
    <xsl:apply-templates/>
  </xsl:copy>
</xsl:template>
<xsl:template mode="sch-only" match="sch:*">
  <xsl:element name="{local-name(.)}" namespace="{namespace-uri(.)}">
    <xsl:copy-of select="@*"/>
    <xsl:apply-templates/>
  </xsl:element>
</xsl:template>

<!--========================================================================-->
<!--infinite loop check for include tree-->

<xsl:template name="check-infinite-loops">
  <xsl:variable name="bad-loop">
    <!--generate any messages-->
    <xsl:call-template name="check-given-include-tree"/>
  </xsl:variable>

  <!--exit if messages-->
  <xsl:if test="string($bad-loop)">
    <xsl:message terminate="yes">
      <xsl:copy-of select="$bad-loop"/>
    </xsl:message>
  </xsl:if>
</xsl:template>

<!--recurse through nested include directives-->
<xsl:template name="check-given-include-tree">
  <xsl:param name="root-node-ids" select="' '"/>
  <xsl:param name="uri" select="'{invocation document}'"/>
  <xsl:param name="last-uri"/>
  <xsl:variable name="this-root-node-id" select="generate-id(document('',/))"/>
  <xsl:choose>
    <xsl:when test="contains( $root-node-ids,
                              concat(' ',$this-root-node-id,' ') )">
      <!--we've ended up pointing to somewhere in our ancestry-->
      <xsl:text>Infinite loop of include statements triggered by </xsl:text>
      <xsl:text/>including '<xsl:value-of select="$uri"/>' from '<xsl:text/>
      <xsl:value-of select="$last-uri"/>'<xsl:text>
</xsl:text>
    </xsl:when>
    <xsl:otherwise>
      <!--so far so good, now check down next level of children-->
      <xsl:variable name="past-root-node-ids"
                   select="concat( $root-node-ids, $this-root-node-id, ' ' )"/>
      <xsl:for-each select="document('',/)/*/vlc:Include">
        <xsl:variable name="new-uri" select="@uri"/>
        <xsl:choose>
          <xsl:when test="not(normalize-space(@uri))">
   <xsl:text>Include directive cannot have an empty uri= attribute: </xsl:text>
            <xsl:call-template name="report-current-node-context"/>
            <xsl:text>
</xsl:text>
          </xsl:when>
          <xsl:otherwise>
            <xsl:for-each select="document($new-uri,.)">
              <xsl:call-template name="check-given-include-tree">
                <xsl:with-param name="root-node-ids" 
                                select="$past-root-node-ids"/>
                <xsl:with-param name="uri" select="$new-uri"/>
                <xsl:with-param name="last-uri" select="$uri"/>
              </xsl:call-template>
            </xsl:for-each>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:for-each>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--========================================================================-->
<!--consistency checks for user's expression of constraints-->

<xsl:template name="check-inconsistencies">
  <xsl:variable name="inconsistencies">
    <!--good refs-->
    <xsl:for-each select="/*/vlc:Contexts/
                          vlc:Context[normalize-space(@values)='']">
      <xsl:text>A context cannot have no references to values: </xsl:text>
      <xsl:call-template name="report-current-node-context"/>
      <xsl:text>
</xsl:text>
    </xsl:for-each>

    <!--defined ids-->
    <xsl:for-each select="/*/vlc:ValueLists/
                          vlc:ValueList[normalize-space(@xml:id)='']">
      <xsl:text>A code list cannot be listed without an xml:id </xsl:text>
      <xsl:text>unique identifier: </xsl:text>
      <xsl:call-template name="report-current-node-context"/>
      <xsl:text>
</xsl:text>
    </xsl:for-each>

    <!--no duplicate ids-->
    <xsl:for-each select="/*/vlc:ValueLists/
                         vlc:ValueList[count(key('valuelists',@xml:id))>1][1]">
    <xsl:text>A code list xml:id identifier must be unique (xml:id="</xsl:text>
      <xsl:value-of select="@xml:id"/>
      <xsl:text>"): 
</xsl:text>
      <xsl:for-each select="key('valuelists',@xml:id)">
        <xsl:text>  </xsl:text>
        <xsl:call-template name="report-current-node-context"/>
        <xsl:text>
</xsl:text>
      </xsl:for-each>
    </xsl:for-each>

    <!--no missing or conflicting contexts-->
    <xsl:for-each select="/*/vlc:Contexts/vlc:Context">
      <xsl:choose>
        <xsl:when test="@xpath and @context">
          <xsl:text>A context cannot have both @xpath and @context: </xsl:text>
          <xsl:call-template name="report-current-node-context"/>
          <xsl:text>
</xsl:text>
        </xsl:when>
      </xsl:choose>
    </xsl:for-each>

    <!--check all document references to point to genericode instances-->
    <xsl:for-each select="/*/vlc:ValueLists/vlc:ValueList">
      <xsl:variable name="doc" select="document(@uri,.)"/>
      <xsl:choose>
        <xsl:when test="not($doc)">
        <xsl:text>Resource not found or cannot be processed as XML: </xsl:text>
          <xsl:value-of select="concat(@uri,': ')"/>
          <xsl:call-template name="report-current-node-context"/>
          <xsl:text>
</xsl:text>
        </xsl:when>
        <xsl:otherwise>
          <xsl:call-template name="code-list-confirm-instance">
            <xsl:with-param name="valuelist" select="."/>
            <xsl:with-param name="doc" select="$doc"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>

  </xsl:variable>

  <xsl:if test="string($inconsistencies)">
    <xsl:message terminate="yes">
      <xsl:value-of select="$inconsistencies"/>
    </xsl:message>
  </xsl:if>
</xsl:template>

<!--========================================================================-->
<!--string manipulation for comparisons-->

<!--most characters not allowed in a part of an attribute or element name (this
    is not meant to be rigourous, just strong enough for obvious errors);
    assumes the axis is removed and the prefix and name are separate-->
<xsl:variable name="forbidden"
  >@:`~!#$%^&amp;*()+=[]{}\|;"'&lt;>?/</xsl:variable>

<xsl:template name="build-metadata-comparison-string">
  <xsl:param name="item"/>
  <xsl:param name="list"/>
  <!--get rid of unwanted space-->
  <xsl:variable name="normitem" select="normalize-space($item)"/>
  <!--determine if an attribute is being addressed-->
  <xsl:variable name="attraxis">
    <xsl:choose>
      <xsl:when test="starts-with($normitem,'@')">@</xsl:when>
      <xsl:when test="starts-with($normitem,'attribute::')"
                                                        >attribute::</xsl:when>
    </xsl:choose>
  </xsl:variable>
  <!--true/false the addressed item is an attribute-->
  <xsl:variable name="attr" select="boolean(string($attraxis))"/>
  <!--determine the qualified name of the item-->
  <xsl:variable name="qname" select="substring-after($normitem,$attraxis)"/>
  <!--determine the prefix of the item name-->
  <xsl:variable name="prefix" select="substring-before($qname,':')"/>
  <!--determine the local part of the item name-->
  <xsl:variable name="name">
    <xsl:choose>
      <xsl:when test="contains($qname,':')">
        <xsl:value-of select="substring-after($qname,':')"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$qname"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!--determine the namespace of the item-->
  <xsl:variable name="ns">
    <!--don't look if the attribute doesn't have a prefix-->
    <xsl:if test="not( $attr and not( $prefix ) )">
      <xsl:value-of select="$item/../namespace::*[name(.)=$prefix]"/>
    </xsl:if>
  </xsl:variable>

  <!--check for any forbidden characters in the item attribute value-->
  <xsl:if test="translate($name,translate($name,$forbidden,''),'') or
                translate($prefix,translate($prefix,$forbidden,''),'')">
    <xsl:message terminate="yes">
      <xsl:text>Unexpected character(s) in item= attribute '</xsl:text>
      <xsl:value-of select="$normitem"/>
      <xsl:text>': </xsl:text>
      <xsl:call-template name="report-current-node-context"/>
    </xsl:message>
  </xsl:if>

  <!--build a string of expressions checking for meta data-->
  <xsl:variable name="properties">
    <xsl:call-template name="document-model-metadata">
      <xsl:with-param name="item" select="$normitem"/>
      <xsl:with-param name="list" select="$list"/>
      <xsl:with-param name="attr" select="$attr"/>
      <xsl:with-param name="name" select="string($name)"/>
      <xsl:with-param name="ns"   select="string($ns)"/>
      <xsl:with-param name="node" select="$item"/>
    </xsl:call-template>
  </xsl:variable>

  <!--the first " and " in the string is not required, so trim it off-->
  <xsl:value-of select="substring($properties,6)"/>
</xsl:template>

<!--parameters:

     $property - the name of the abstract code list meta data property (not the
                 name of the meta data element in the code list file)
     $list     - <valueList> node from the association
     $itemattr - true/false that the item being tested is an attribute
     $metaattr - true/false that the meta data value is an attribute
     $metans   - namespace of the meta data element/attribute
     $metaname - name of the meta data element/attribute
     $address  - (deprecated) only use an explicit XPath address if the other
                 parameters are not sufficient to describe it in the abstract;
                 any value here will cause the other parameters to be ignored
-->
<xsl:template name="check-list-property">
  <xsl:param name="property"/>
  <xsl:param name="list"/>
  <xsl:param name="itemattr" select="true()"/>
  <xsl:param name="metaattr" select="true()"/>
  <xsl:param name="metans"/>
  <xsl:param name="metaname"/>
  <xsl:param name="address">
    <!--when not supplied, must compose an address from the other arguments-->
    <xsl:choose>
      <xsl:when test="not($metaattr)">
        <!--the meta data is an element, must be item's parent's child-->
        <xsl:text>../</xsl:text>
      </xsl:when>
      <!--the meta data is an attribute-->
      <xsl:when test="$itemattr">
        <!--the meta data is a sibling attribute of the item attribute-->
        <xsl:text>../@</xsl:text>
      </xsl:when>
      <xsl:otherwise>
        <!--the meta data is an attached attribute of the item element-->
        <xsl:text>@</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
    <!--the axis is done, just name the element or attribute now-->
    <xsl:choose>
      <xsl:when test="$metaattr and not(normalize-space($metans))">
        <!--a non-namespaced attribute is simply addressed by name-->
        <xsl:value-of select="$metaname"/>
      </xsl:when>
      <xsl:otherwise>
        <!--elements and namespaced attributes need uri checking-->
        <xsl:text>*[namespace-uri(.)='</xsl:text>
        <xsl:value-of select="$metans"/>
        <xsl:text>' and local-name(.)='</xsl:text>
        <xsl:value-of select="$metaname"/>
        <xsl:text>']</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:param>
  <!--determine the meta data from either the masquerade or external file-->
  <xsl:variable name="code-list-property-value">
    <xsl:choose>
      <!--first check for masquerading meta data-->
      <xsl:when test="$property='Reference' and 
                      $list/vlc:MetaData/vlc:Reference">
        <xsl:value-of select="$list/vlc:MetaData/vlc:Reference"/>
      </xsl:when>
      <xsl:when test="$property='Name' and 
                      $list/vlc:MetaData/vlc:Name">
        <xsl:value-of select="$list/vlc:MetaData/vlc:Name"/>
      </xsl:when>
      <xsl:when test="$property='ID' and 
                      $list/vlc:MetaData/vlc:ID">
        <xsl:value-of select="$list/vlc:MetaData/vlc:ID"/>
      </xsl:when>
      <xsl:when test="$property='URI' and 
                      $list/vlc:MetaData/vlc:URI">
        <xsl:value-of select="$list/vlc:MetaData/vlc:URI"/>
      </xsl:when>
      <xsl:when test="$property='Version' and 
                      $list/vlc:MetaData/vlc:Version">
        <xsl:value-of select="$list/vlc:MetaData/vlc:Version"/>
      </xsl:when>
      <xsl:when test="$property='VersionURI' and 
                      $list/vlc:MetaData/vlc:VersionURI">
        <xsl:value-of select="$list/vlc:MetaData/vlc:VersionURI"/>
      </xsl:when>
      <xsl:when test="$property='LocationURI' and 
                      $list/vlc:MetaData/vlc:LocationURI">
        <xsl:value-of select="$list/vlc:MetaData/vlc:LocationURI"/>
      </xsl:when>
      <xsl:when test="$property='AgencyName' and 
                      $list/vlc:MetaData/vlc:AgencyName">
        <xsl:value-of select="$list/vlc:MetaData/vlc:AgencyName"/>
      </xsl:when>
      <xsl:when test="$property='AgencyID' and 
                      $list/vlc:MetaData/vlc:AgencyID">
        <xsl:value-of select="$list/vlc:MetaData/vlc:AgencyID"/>
      </xsl:when>
      <xsl:otherwise>
        <!--need to go to the external file for the meta data string-->
        <xsl:call-template name="code-list-get-property-value">
          <xsl:with-param name="property" select="$property"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:variable>
  <!--express the boolean expression of a property's value needing to be set;
      note by using a double negative of not(x!=y) instead of (x=y) is required
      so that the test will not fail if an operand is not present.  Using a
      double negative will return true if the operand is not present, and will
      return false if the operand is present but the wrong value.-->
  <xsl:if test="string($code-list-property-value)">
    <xsl:value-of select='concat(" and not(",string($address),
                          "!=&apos;",$code-list-property-value,"&apos;)")'/>
  </xsl:if>
</xsl:template>

<!--========================================================================-->
<!--diagnostic report-->

<xsl:template name="report-current-node-context">
  <!--start with the XPath address of the element in the context expression-->
  <xsl:for-each select="ancestor-or-self::*">
    <xsl:text/>/<xsl:value-of select="name(.)"/>
    <xsl:variable name="preceding-count" 
                select="count(preceding-sibling::*[name(.)=name(current())])"/>
    <xsl:if test="$preceding-count > 0">
      <xsl:text/>[<xsl:value-of select="$preceding-count + 1"/>]<xsl:text/>
    </xsl:if>
  </xsl:for-each>
  <!--suffix with attribute or namespace node as appropriate-->
  <xsl:if test="not(self::*)">
    <xsl:text>/</xsl:text>
    <xsl:if test="count(../@*) = count(.|../@*)">@</xsl:if>
    <xsl:if test="count(../namespace::*) = count(.|../namespace::*)">
      <xsl:text>namespace::</xsl:text>
    </xsl:if>
    <xsl:value-of select="name(.)"/>
  </xsl:if>
</xsl:template>

</xsl:stylesheet>